//----------------------------------------------------------------------------
// 
//	file: spike_layer.h
// 
//	function: spiking lissom layer 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Sat Aug  1 03:31:31 CDT 1998
// 
//	$Revision: 1.7 $ $Date: 2001/11/30 20:24:15 $
//	$Log: spikelayer.h,v $
//	Revision 1.7  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.6  2000/02/08 11:25:21  yschoe
//	*** empty log message ***
//
//	Revision 1.5  1998/11/17 12:44:56  yschoe
//	1. fixed activation to exactly match gaus-slissom 97fall
//	2. added print_stat() for neuron stat gathering
//
//	Revision 1.4  1998/10/26 03:58:20  yschoe
//	1. relocated reset_matrix() to appropriate positions where it is
//	   immediately necessary
//	2. calc_spike() moved to the beginning of the activate() function,
//	   as it was done in gaus-slissom/98jun25/
//	3. output = avg_spiking in learn(). This was necessary so that it is
//	   the same as gaus-slissom/98jun25
//
//	Revision 1.3  1998/10/16 10:08:04  yschoe
//	added force_match(char *token) and made match funcitons to return
//	0 or 1 (fail/success).
//
//	Revision 1.2  1998/10/05 19:48:46  yschoe
//	working for the first time
//
//	Revision 1.1  1998/09/04 03:36:37  yschoe
//	fixed all warnings generated by -Wall
//
//
//----------------------------------------------------------------------------
#ifndef __SPIKE_LAYER_H
#define __SPIKE_LAYER_H

#include "spike-code.h"
#include "netlayer.h"
#include "real.h"
#include "ppmio.h"

// spike queue length
// #define SPIKE_QLEN	3 : was 8
#define SPIKE_QLEN 16
// spike queue mask : 0x07 == 0 0111(binary) : 3 bits
// #define SPIKE_QMASK	0x07
#define SPIKE_QMASK 0xffff
// history queue length
#define SPIKE_HISTQLEN  16
// history queue mask : 0x3FF == 11 1111 1111(binary) : 10 bits
#define SPIKE_HISTQMASK	0xFFFF

typedef int SPIKE_Q;

//----------------------------------------------------------------------------
//	CLASS spikelayer
//	  - spiking lissom layer - derived from netlayer
//----------------------------------------------------------------------------
class spikelayer : public netlayer {

  // General idea:
  //	Put decayed spike activity in this->output[x][y], so that
  //	it could be generically used with other types of netlayers.
  //	No spikes actually goes through the connection this way.
  //	Thus the postsynaptic potential is calculated for all 
  //	targets receiving from a certain neuron.

public:

  int      **spike;

  // added 00jun6 ----------------------------------------------
  REAL	   **new_s_exc; // Accumulated outgoing spikes 
  REAL	   **new_s_inh; //
  REAL	   **new_s_int; //
  REAL	   **s_exc;	// 
  REAL	   **s_inh;	//
  REAL	   **s_int;	//

  REAL	   **new_eta;	// activity (eta) -- compared to thresh
  REAL	   **eta;	//
  REAL	   init_eta_randomness; // initialize eta randomly

  REAL	   **new_thresh; // dynamic threshold (theta)
  REAL	   **thresh;	 //

  REAL	   **v;		// membrand voltage
  REAL	   **r;		// refractory period
  REAL	   **new_r;		// refractory period
  REAL 	   hard_refractory;
  // end of addition 00jun6 -------------------------------------

  SPIKE_Q  **spikeact_q;   // spike activity (MSB: oldest, LSB: newest spike)
			   // doubles as synapse Q for outgoing connecions
  SPIKE_Q  **inhloop_q;   // inhibitory loop in spike generator
  REAL	   **avg_spiking;  // avg spiking rate
  REAL	   **lfp;	  // act_function(net_input) -- for display

  int	   auto_theta_flag;
  REAL	   auto_theta_rate;
protected:

  // decay rate for incoming connections and spike coder
  REAL	   synapse_decay;
  REAL	   inhloop_decay;

public:
  // spike (base) threshold
  REAL	   theta;
  // contribution of inhibitory feedback in threshold calculation
  REAL	   tau;

protected:

  // Queue sum look up tables
  REAL	   *synapse_lut; // synapse decay rate
  REAL	   *inhloop_lut;
  REAL	   *bitcount_lut; // quickly return bits in a byte (or 2 bytes)

private:

  inline void push_spike(int *queue, int spike);

  inline REAL norm_factor(REAL sum) { // aff normalization factor
        return sqrt(sum);
        //return sum;
  };
  inline REAL norm_sum(REAL sum) { // Used when summing afferent wgts
        return sum*sum;                    //  for normalization.
        //return sum;
  };


public:

	spikelayer(char *name, int cols, int rows, int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
                 int inp_cols, int inp_rows, int  aff_rad,
		 REAL synapse_decay, REAL inhloop_decay, REAL theta=0.0, 
		 REAL tau=0.0,
		 int padded_flag=1);
       ~spikelayer();

	// calc_mua(REAL **input);  // calculate mua group given an input
	// calc_correl(int area1, int area2);// calculate auto and cross-correl

  virtual void activate(int calc_aff = 0);
  virtual void calc_spike();
  virtual void push_spike();

  // overload previous definition in class netlayer
  virtual void learn();

  // print status of (col,row) neuron 
  void print_stat(int col, int row);
  void print_stat(int cur_inp, int inp_col, int inp_row);

  void clear_activity() { reset_matrix(output,cols,rows); 
		reset_matrix(spike,cols,rows);
		reset_matrix(avg_spiking,cols,rows);
		reset_matrix(spikeact_q,cols,rows);
		reset_matrix(inhloop_q,cols,rows);
		}

  // adjust sigmoid values to make activity more well-behaved
  void adjust_sigmoid(int col, int row, REAL percent_active, REAL contrast);

};

#endif
